from django.db import migrations
from django_native_search.fields import OccurrencesField

class OccurrencesAddPrefixField(migrations.AlterField):
    def __init__(self, model_name):
        super().__init__(model_name=model_name,
            name='occurrences',
            field=OccurrencesField(query_name="occurrence"))
        
    def database_forwards(self, app_label, schema_editor, from_state, to_state):
        from_model=from_state.apps.get_model(app_label, self.model_name)
        from_field = from_model._meta.get_field(self.name)
        through=from_field.remote_field.through
        schema_editor.add_field(through, through._meta.get_field('prefix'))



from django.db import models
from django.db.models.signals import post_save

class OccurrencesField(models.ManyToManyField):
    def __init__(self, query_name=None, **kwargs):
        from .models import Lexem
        kwargs.setdefault('to', Lexem._meta.label)
        kwargs.setdefault('related_name','+')
        kwargs.setdefault('editable', False)
        super().__init__(**kwargs)
        self.query_name=query_name
        
    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        kwargs["query_name"]=self.query_name
        return name, path, args, kwargs
        
    def contribute_to_class(self, cls, name, **kwargs):
        super().contribute_to_class(cls, name, **kwargs)
        occurrences = self.remote_field.through
        if not occurrences or isinstance(occurrences, str):
            return
        
        models.PositiveIntegerField(db_index=True).contribute_to_class(
            occurrences, 'position')
        
        models.CharField(db_index=True, max_length=16, default=' ').contribute_to_class(
            occurrences, 'prefix')
        
        unique = occurrences._meta.unique_together[0]+('position',)
        occurrences._meta.unique_together=(unique,)
        occurrences._meta.ordering=['position']
        occurrences._meta.verbose_name=f"{cls._meta.model_name}-occurrence"
        occurrences._meta.verbose_name_plural=f"{cls._meta.model_name}-occurrences"
        remote_field=occurrences._meta.get_field(cls._meta.model_name).remote_field
        remote_field.related_query_name=self.query_name
        
        class RelatedAccessor(remote_field.field.related_accessor_class):
            
            def __get__(self, instance, cls=None):
                if instance:
                    return super().__get__(instance, cls)
                
                return self.rel.related_model._meta.default_manager
        
        setattr(cls, self.name, RelatedAccessor(remote_field))
        
        post_save.connect(self.update_occurrences)
        
    def update_occurrences(self, instance, **kwargs):
        if not isinstance(instance, self.model):
            return
        lexem_model=self.remote_field.model
        max_length=lexem_model.surface.field.max_length
        instance.occurrences.all().delete()
        
        lexem_cache={}
        for position, surface in enumerate(instance.tokens):
            if len(surface)>max_length:
                continue
            
            lexem=lexem_cache.get(surface) or lexem_cache.setdefault(
                surface, lexem_model.objects.get_or_create(surface=surface)[0])
            values=dict(lexem=lexem, position=position)
            prefix=getattr(surface,'prefix',None)
            if prefix is not None:
                values['prefix']=prefix
            instance.occurrences.create(**values)



# Generated by Django 3.0.8 on 2020-07-22 21:10

from django.db import migrations, models
import django.db.models.deletion
import django.db.models.functions.text
import django_expression_index.index


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Lexem',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('surface', models.CharField(db_index=True, max_length=255, unique=True)),
            ],
        ),
        migrations.CreateModel(
            name='LexemTail',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('surface', models.CharField(db_index=True, max_length=255)),
                ('lexem', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='tails', related_query_name='tail', to='django_native_search.Lexem')),
            ],
        ),
        migrations.AddIndex(
            model_name='lexem',
            index=django_expression_index.index.ExpressionIndex(expressions=[django.db.models.functions.text.Lower('surface')], name='django_native_searc_9a3539_idx'),
        ),
        migrations.AddIndex(
            model_name='lexemtail',
            index=django_expression_index.index.ExpressionIndex(expressions=[django.db.models.functions.text.Lower('surface')], name='django_native_searc_87adea_idx'),
        ),
        migrations.AlterUniqueTogether(
            name='lexemtail',
            unique_together={('lexem', 'surface')},
        ),
    ]



# Generated by Django 3.0.8 on 2020-08-25 13:12

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('django_native_search', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Index',
            fields=[
            ],
            options={
                'verbose_name_plural': 'indexes',
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('contenttypes.contenttype',),
        ),
    ]



from functools import partial, cache

from django import forms
from django.views.generic.edit import FormMixin
from django.utils.text import capfirst
from django.core.exceptions import ValidationError

ALL_VALUE = "all"

class SearchForm(forms.Form):
    q = forms.CharField(
        required=True,
        label="Query",
        widget=forms.TextInput(attrs={"type": "search", 'style': "width:400px"}),
    )
    
    def clean(self):
        for key in list(self.cleaned_data):
            if key not in self.data:
                del self.cleaned_data[key]
        return super().clean()
    
    def clean_q(self):
        query = self.cleaned_data['q']
        tokens=self.index.tokenize(query)
        try:
            next(tokens)
        except StopIteration:
            raise ValidationError(f"'{query}' does not contain relevant keywords.")
        return query
    
    def search(self):
        filters = dict([(key if isinstance(val, str) else f"{key}__in", val) 
                        for key, val in self.cleaned_data.items()
                        if key != 'q' and val != ALL_VALUE])
        
        return self.index.objects.filter(**filters).search(self.cleaned_data["q"])

def field_values(field):
    return field.model.objects.all().values_list(field.name, flat=True).distinct()

def field_choice_gen(field, empty_label, all_label):
    if all_label:
        yield ALL_VALUE, all_label
    for item in field_values(field):
        yield item, item if item else empty_label

def field_choices(field, empty_label, all_label):
    return list(field_choice_gen(field, empty_label, all_label))

def searchform_choice_field(field, formfield_class, 
                           empty_label="---------", all_label="All", 
                           widget_attrs={}, **kwargs):
    attrs = dict(label=capfirst(field.verbose_name),
                 widget = formfield_class.widget(**widget_attrs), 
                 choices = partial(field_choices, field, empty_label, all_label), 
                 required = False)
    
    if all_label:
        attrs.setdefault("initial", ALL_VALUE)
    attrs.update(**kwargs)
    return formfield_class(**attrs)

def searchform_single_choice_field(field, **kwargs):
    return searchform_choice_field(field, forms.TypedChoiceField, **kwargs)

def searchform_multiple_choice_field(field, **kwargs):
    kwargs.setdefault("all_label", None)
    kwargs["empty_value"] = None
    return searchform_choice_field(field, forms.TypedMultipleChoiceField, **kwargs)


def searchform_factory(index, base=SearchForm, 
                       formfield_callback=searchform_multiple_choice_field,
                       formfield_attrs={}):
    class_name = index.__name__ + 'SearchForm'
    attrs={
        'index':index
        }
    for field in index._meta.fields:
        if field.db_index:
            if formfield_callback:
                formfield = formfield_callback(field, **formfield_attrs)
                if formfield:
                    attrs[field.name]= formfield
            else: 
                attrs[field.name]=field.formfield(**formfield_attrs)
    return type(class_name, (base,), attrs)

class GetFormMixin(FormMixin):
    def get(self, request, *args, **kwargs):
        form=self.get_form()
        if not form.is_bound:
            return super().get(request, *args, form=form, **kwargs)
        if form.is_valid():
            return self.form_valid(form)
        else:
            return self.form_invalid(form)
    
    @cache
    def get_form(self):
        form = super().get_form()
        for f in form.fields.values():
            if hasattr(f,'choices'):
                f.choices=f.choices
        return form
    
    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        if self.request.GET:
            kwargs["data"]=self.request.GET
        return kwargs

class SearchFormMixin(GetFormMixin):
    results_per_page=25
    
    def get_context_data(self, form, **kwargs):
        context = super().get_context_data(**kwargs)
        if form.is_valid():
            context['results'] = form.search().prefetch_matches()
        return context
    
    def form_valid(self, form):
        return self.render_to_response(self.get_context_data(form=form))



import copy
from django.core.exceptions import FieldDoesNotExist
from django.db.models import (F, Value, Min, OuterRef, Count, FloatField, QuerySet, Q, Prefetch,
                              ExpressionWrapper)
from django.db.models.manager import BaseManager, Manager
from django.db.models.functions import Abs
from django.conf import settings
import logging
from functools import cache



MAX_RANKING_KEYWORDS_COUNT=getattr(settings,"SEARCH_MAX_RANKING_KEYWORDS_COUNT", 3)

logger=logging.getLogger(__name__)


def prefix_lookups(q, prefix):
    if isinstance(q, tuple):
        return prefix+q[0], q[1]
    for i,c in enumerate(q.children):
        q.children[i]=prefix_lookups(c, prefix)
    return q


class SearchQuerySet(QuerySet):
    def __init__(self, model=None, query=None, using=None, hints=None):
        super().__init__(model=model, query=query, using=using, hints=hints)
        self.search_conditions=[]
    
    @cache
    def count(self):
        return self.values('pk').aggregate(c=Count("*"))['c']
        
    def apply_filter(self, q):
        filtered = self.filter(prefix_lookups(copy.deepcopy(q), "occurrence__"))
        filtered.search_conditions.append(q)
        return filtered
    
    def search(self, query):
        ranking=self
        filtered=self
        conditions=self.model.parse_query(query)
        sticked=False
        for q in conditions:
            ranking=ranking.apply_filter(q).annotate_rank()
            if getattr(q,'sticky', None):
                ranking=ranking.filter(d=1)
                sticked=True
            filtered=self.apply_filter(q).filter(pk__in=filtered.all())
        
        if filtered is self:
            return self

        if sticked:
            filtered=filtered.filter(pk__in=ranking)

        results = self.filter(pk__in=filtered)
        results = results.annotate(rank=ranking.filter(pk=OuterRef("pk")).values("rank")).order_by("rank")
        results.search_conditions=conditions
        return results
    
    def annotate_rank(self):
        ranking=self
        
        keycount = len(self.search_conditions)
        
        if keycount==1:
            ranking=ranking.annotate(dsum=Value(1, output_field=FloatField()))
        else:
            ranking=ranking.annotate(d=F("occurrence__position")-F("p"))
            ranking=ranking.annotate(dsum=Abs(F('d')-1.0, output_field=FloatField())+F("dsum"))
        
        if keycount<=MAX_RANKING_KEYWORDS_COUNT:
            ranking=ranking.annotate(
                    rank=ExpressionWrapper(Min("dsum")*F("length")/Count("*"), output_field=FloatField()))

        ranking=ranking.annotate(p=F("occurrence__position"))
        return ranking
    
    def prefetch_matches(self):
        qs=self.model.occurrences.filter(Q(*self.search_conditions, _connector=Q.OR))
        return self.prefetch_related(Prefetch("occurrences", 
                                  queryset=qs,
                                  to_attr="matches"))
    
    def _clone(self):
        c = super()._clone()
        c.search_conditions=self.search_conditions[:]
        return c

class IndexManager(Manager):
    root_indexentry_models=[]
    
    def get_queryset(self):
        qs=super().get_queryset()
        return qs.filter(Q(*[Q(app_label=model._meta.app_label, model = model._meta.model_name)
                            for model in self.indexentry_models], _connector=Q.OR))

    @property
    def indexentry_models(self):
        for model in self.root_indexentry_models:
            for m in self.iter_descendants(model):
                yield m
    
    def iter_descendants(self, cls):
        yield cls
        for child in cls.__subclasses__():
            for m in self.iter_descendants(child):
                yield m
    
    @classmethod
    def register(cls, model):
        cls.root_indexentry_models.append(model)


class IndexEntryManager(BaseManager.from_queryset(SearchQuerySet)):
    def contribute_to_class(self, cls, name, **kwargs):
        super().contribute_to_class(cls, name, **kwargs)
        IndexManager.register(cls)

    @property
    def target_model(self):
        try:
            return self.model._meta.get_field(self.model.object_field).remote_field.model
        except FieldDoesNotExist:
            return None
        
    def get_or_prepare(self, obj):
        model=self.get_index_model(obj._meta.model)
        if not model:
            raise RuntimeError(f"Index for {obj._meta.model_name} is not configured.")
        try:
            indexed=model._meta.default_manager.get(**{model.object_field:obj})
            setattr(indexed,model.object_field, obj)
            return indexed
        except model.DoesNotExist:
            return model(object=obj)
    
    def get_index_model(self, model):
        if model == self.target_model:
            return self.model
        
        for subcls in self.model.__subclasses__():
            index = subcls._meta.default_manager.get_index_model(model)
            if index:
                return index
        return None
    
    def refresh(self, objects, break_on_failure=False):
        for obj in objects:
            try:
                self.get_or_prepare(obj).save()
            except Exception:
                logger.exception(f"Exception raised when updating index for '{obj}'")
                if break_on_failure:
                    raise
    
    def rebuild(self):
        if self.target_model:
            return self.refresh(self.model.get_index_queryset())
            
        for subcls in self.model.__subclasses__():
            subcls._meta.default_manager.rebuild()
            
    



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    author='Kamil Mierzejewski',
    name='django-native-search',
    version='0.5.4',
    description='A simple search engine using native django database backend.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/kmierzeje/django-native-search',
    packages=setuptools.find_packages(),
    install_requires=['django>=3.0.8', 'django-expression-index>=0.1.0'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.7',
    ],
)



from django.contrib import admin
from django.contrib.humanize.templatetags.humanize import intcomma
from .models import Lexem, Index

@admin.register(Lexem)
class LexemAdmin(admin.ModelAdmin):
    list_display=['surface']
    search_fields=['surface']
    

@admin.register(Index)
class IndexAdmin(admin.ModelAdmin):
    readonly_fields = ['app_label', 'model', 'entries', 'occurrences']
    list_display = ['__str__', 'entries', 'occurrences']
    
    def has_add_permission(self, request):
        return False
    
    def has_change_permission(self, request, obj=None):
        return False
    
    def has_delete_permission(self, request, obj=None):
        return False
    
    def entries(self, obj):
        return intcomma(obj.entries())
    def occurrences(self, obj):
        return intcomma(obj.occurrences())
        


import logging
import re
from django.db import models
from django.db.models.functions import Lower
import django_expression_index

from django.template.loader import render_to_string
from django.utils.functional import cached_property
from django.db.models.signals import post_save
from django.dispatch import receiver

from .manager import IndexEntryManager, IndexManager
from django.utils.safestring import mark_safe
from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django_native_search.fields import OccurrencesField


MIN_TAIL_LEN=getattr(settings,"SEARCH_MIN_SUBSTR_LENGTH", 2)
MAX_TAIL_COUNT_IN_QUERY=getattr(settings, "SEARCH_MAX_SUBTSTR_COUNT_IN_QUERY", 300)
MAX_EXCERPT_FRAGMENTS=getattr(settings, "SEARCH_MAX_EXCERPT_FRAGMENTS", 5)
EXCERPT_FRAGMENT_START_OFFSET=getattr(settings, "SEARCH_EXCERPT_FRAGMENT_START_OFFSET", -3)
EXCERPT_FRAGMENT_END_OFFSET=getattr(settings, "SEARCH_EXCERPT_FRAGMENT_END_OFFSET", 6)
EXCERPT_ADDITONAL_CONTEXT_FACTOR=getattr(settings, "SEARCH_EXCERPT_ADDITONAL_CONTEXT_FACTOR", 2)

logger=logging.getLogger(__name__)

class Lexem(models.Model):
    surface=models.CharField(max_length=255, db_index=True, unique=True)
    
    class Meta:
        indexes=[django_expression_index.ExpressionIndex(expressions=[Lower('surface')])]
    
    def __str__(self):
        return self.surface

class LexemTail(models.Model):
    lexem=models.ForeignKey(Lexem, on_delete=models.CASCADE, 
                            related_name="tails", related_query_name='tail')
    surface=models.CharField(max_length=255, db_index=True)
    
    class Meta:
        indexes=[django_expression_index.ExpressionIndex(expressions=[Lower('surface')])]
        unique_together=('lexem','surface')
    
    def __str__(self):
        return self.surface

@receiver(post_save, sender=Lexem)
def update_lexem_tail(instance, **kwargs):
    instance.tails.all().delete()
    for i in range(len(instance.surface)):
        tail=instance.surface[i:]
        if len(tail)>MIN_TAIL_LEN:
            instance.tails.create(surface=tail)


models.CharField.register_lookup(Lower)

class Token(str):
    pass

class IndexEntry(models.Model):
    length=models.PositiveIntegerField(editable=False)
    occurrences=OccurrencesField(query_name="occurrence")
    
    object_field="object"
    objects=IndexEntryManager()

    search_template=None
    
    class Meta:
        abstract=True
        
    
    def save(self, force_insert=False, force_update=False, using=None, 
        update_fields=None):
        logger.info(f"Indexing {self}...")
        self.length=len(self.tokens)
        super().save(force_insert=force_insert, 
                            force_update=force_update, 
                            using=using, 
                            update_fields=update_fields)
    
    @cached_property
    def tokens(self):
        return list(self.prepare_text())
    
    def prepare_text(self):
        return self.tokenize(self.rendered_text)
    
    token_pattern = re.compile(r'[^\s"]+')
    quote = '"'
    
    @classmethod
    def tokenize(cls, text):
        i=0
        sticky=False
        while True:
            res=cls.token_pattern.search(text, i)
            if not res:
                break
            
            token = Token(res.group(0))
            token.prefix=re.sub(r"\s+"," ",text[i:res.start()])
            if cls.quote:
                quotes=token.prefix.count(cls.quote)
                if sticky and quotes>0:
                    sticky = False
                    quotes-=1
                
                token.sticky=sticky
                if quotes%2>0:
                    sticky=not sticky
            i=res.end()
            yield token
    
    @classmethod
    def parse_query(cls, query):
        lookup = 'surface'
        if query.islower():
            lookup +="__lower"
        tokens=list(cls.tokenize(query))
        if len(tokens)==1 and len(tokens[0])>MIN_TAIL_LEN:
            tail_q=LexemTail.objects.filter(**{
                lookup+"__gte":tokens[0],
                lookup+"__lt":tokens[0]+chr(0x10FFFF)})
            if tail_q.count()<=MAX_TAIL_COUNT_IN_QUERY:
                return [models.Q(lexem__in=Lexem.objects.filter(
                    tail__in=tail_q))]
        
        query=[]
        for token in tokens:
            condition=models.Q(lexem__in=Lexem.objects.filter(**{lookup:token}))
            if query:
                condition.sticky=getattr(token,'sticky',False)
            query.append(condition)
        return query
        
    @property
    def excerpt(self):
        matches=getattr(self, 'matches', None)
        if not matches:
            return ""
        
        best_matches={matches[0].lexem_id:(matches[0], 0xFFFF)}
        last_match=matches[0]
        for m in matches[1:]:
            rank=m.position-last_match.position
            last=best_matches.get(last_match.lexem_id)
            if not last or rank<last[1]:
                best_matches[last_match.lexem_id]=last_match, rank
            last_match=m
        
        best_matches=set([m[0] for m in list(best_matches.values())[:MAX_EXCERPT_FRAGMENTS]])
        
        i=1
        while len(best_matches)<MAX_EXCERPT_FRAGMENTS and i<len(matches):
            best_matches.add(matches[i])
            i+=1
            
        additional_context=(MAX_EXCERPT_FRAGMENTS-len(best_matches))*EXCERPT_ADDITONAL_CONTEXT_FACTOR
        
        words=self.occurrences.filter(models.Q(*[
            models.Q(position__gt=match.position+EXCERPT_FRAGMENT_START_OFFSET-additional_context,
                     position__lt=match.position+EXCERPT_FRAGMENT_END_OFFSET+additional_context)
            for match in best_matches], _connector=models.Q.OR))
        
        highlight=set([m.position for m in matches])
        excerpt=""
        pos=-1
        for word in words.select_related('lexem'):
            if word.position>pos+1:
                excerpt+="..."
            if pos>0:
                excerpt+=word.prefix
            
            if word.position in highlight:
                excerpt+= self.highlight(word)
            else:
                excerpt+=word.lexem.surface
            pos=word.position+1
        if pos<self.length:
            excerpt+="..."
        return mark_safe(excerpt)
    
    def highlight(self, word):
        return f"<em>{word.lexem.surface}</em>"
    
    @cached_property
    def rendered_text(self):
        return render_to_string(self.search_template, 
                                {self.object_field:getattr(self, self.object_field)})
        
    @cached_property
    def indexed_text(self):
        return "".join([o.prefix+o.lexem.surface for o in self.occurrences.select_related('lexem')])
        
    @classmethod
    def get_index_queryset(cls):
        return cls.objects.target_model._meta.default_manager.all()
    

class Index(ContentType):
    objects = IndexManager()
    class Meta:
        proxy=True
        verbose_name_plural="indexes"
        
    def occurrences(self):
        return self.model_class().objects.aggregate(models.Count('occurrence'))['occurrence__count']
    
    def entries(self):
        return self.model_class().objects.all().count()
